﻿using System.Collections;
using System.Collections.Generic;
using System;
using Barsalt.ParameterSystem.Exceptions;
using System.Xml.Serialization;
using UnityEngine;

namespace Barsalt.ParameterSystem.Container {
	[Serializable()]
	public class ParameterGroup : IParameterMap, IXmlSerializable {
		#region OBJECT_ATTRIBUTES
		//==========================================================================
		private readonly Dictionary<string, object> PARAMETERS	= new Dictionary<string, object>();
		private readonly Dictionary<string, bool> IS_UPDATED	= new Dictionary<string, bool>();
		//==========================================================================
		#endregion

		#region CONSTRUCTOR
		//==========================================================================
		public ParameterGroup(){}
		//==========================================================================
		#endregion

		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// Uses the key to add a new entry to the group. 
		/// </summary>
		/// <param name="key">The key of the entry</param>
		/// <param name="value">The vale of the entry</param>
		/// <returns>Returns true if the entry was added to the group</returns>
		public bool Add(string key, object value) {
			if(!PARAMETERS.ContainsKey(key)) {
				PARAMETERS.Add(key, value);
				IS_UPDATED.Add(key, true);

				return true;
			} else {
				return false;
			}
		}

		/// <summary>
		/// Uses the key to update the value of the correspoding entry.
		/// </summary>
		/// <param name="key">The key of the entry</param>
		/// <param name="value">The new value for the entry</param>
		/// <returns>Returns true if the value was updated</returns>
		public bool Update(string key, object value){
			if(PARAMETERS.ContainsKey(key)){
				PARAMETERS[key]	= value;
				IS_UPDATED[key]	= true;
			
				return true;
			} else {
				return false;
			}
		}

		/// <summary>
		/// Uses the key to remove an entry from the group.
		/// </summary>
		/// <param name="key">The key of the group</param>
		/// <returns>Returns true if the entry was removed</returns>
		public bool Remove(string key) {
			return (PARAMETERS.Remove(key) && IS_UPDATED.Remove(key));
		}

		/// <summary>
		/// Checks if the group as an entry with the specified key.
		/// </summary>
		/// <param name="key">The specified key</param>
		/// <returns>Returns true if the group has an entry corresponding to the key</returns>
		public bool ContainsKey(string key) {
			return PARAMETERS.ContainsKey(key);
		}

		/// <summary>
		/// Deserializes the ParameterGroup from an XmlReader.
		/// </summary>
		/// <param name="reader">The reader containing the serialized ParameterGroup</param>
		public void ReadXml(System.Xml.XmlReader reader) {
			reader.MoveToContent();

			int count = int.Parse(reader.GetAttribute("count"));

			#region READ_ENTRIES
			//==========================================================================
			reader.ReadToFollowing("Parameter");

			for(int index = 0; index < count; index++) {
				string key = reader.GetAttribute("key");

				XmlSerializer serializer = new XmlSerializer(Type.GetType(reader.GetAttribute("type") + ", " + reader.GetAttribute("assembly")));

				reader.Read();

				object value = serializer.Deserialize(reader);

				PARAMETERS.Add(key, value);
				IS_UPDATED.Add(key, true);

				reader.ReadToFollowing("Parameter");
			}
			//==========================================================================
			#endregion
		}

		/// <summary>
		/// Uses the XmlWriter to serialize the ParameterGroup into xml-format. 
		/// </summary>
		/// <param name="writer">The writer for the serialization</param>
		public void WriteXml(System.Xml.XmlWriter writer) {
			writer.WriteAttributeString("count", PARAMETERS.Count.ToString());

			#region WRITE_ENTRIES
			//==========================================================================
			foreach(string key in PARAMETERS.Keys) {
				writer.WriteStartElement("Parameter");
				writer.WriteAttributeString("key", key);
				writer.WriteAttributeString("type", PARAMETERS[key].GetType().ToString());
				writer.WriteAttributeString("assembly", PARAMETERS[key].GetType().Assembly.ToString());

				XmlSerializer serializer = new XmlSerializer(PARAMETERS[key].GetType());

				serializer.Serialize(writer, PARAMETERS[key]);

				writer.WriteEndElement();
			}
			//==========================================================================
			#endregion
		}

		public List<string> OnDistribution() {
			List<string> keys = new List<string>();

			foreach(string key in PARAMETERS.Keys) {
				if(IS_UPDATED[key]) {
					keys.Add(key);
				}
			}

			foreach(string key in keys) {
				IS_UPDATED[key] = false;
			}

			return keys;
		}
		//==========================================================================
		#endregion

		#region GETTER_AND_SETTER
		//==========================================================================
		public List<string> Keys {
			get {
				List<string> keys = new List<string>();

				foreach(string key in PARAMETERS.Keys){
					keys.Add(key);
				}

				return keys;
			}	
		}

		/// <summary>
		/// Uses the key to retrive a value from the group and returns it as T
		/// </summary>
		/// <typeparam name="T">The Type of the value</typeparam>
		/// <param name="key">The key of the value</param>
		/// <returns>The value</returns>
		public T GetValue<T>(string key) {
			if(PARAMETERS.ContainsKey(key)) {
				object value = PARAMETERS[key];

				if(value.GetType() is T) {
					return (T)value;
				} else {
					throw new PreferenceTypeException(typeof(T), key, value);
				}
			} else {
				throw new ParameterKeyException(key);
			}
		}

		/// <summary>
		/// Uses the specified key to return the datatype of the corresponding value.
		/// </summary>
		/// <param name="key">The specified key</param>
		/// <returns>The Type of the value</returns>
		public Type GetType(string key) {
			if(PARAMETERS.ContainsKey(key)){
				return PARAMETERS[key].GetType();
			}

			return null;
		}

		/// <summary>
		/// Uses the specified key to check if the valus is of Type type. 
		/// </summary>
		/// <param name="key">The specified key</param>
		/// <param name="type">The specified Type</param>
		/// <returns>Returns true if the value is typeof Type</returns>
		public bool IsTypeOf(string key, Type type) {
			if(PARAMETERS.ContainsKey(key) && PARAMETERS[key].GetType() == type){
				return true;
			}

			return false;
		}

		/// <summary>
		///
		/// </summary>
		/// <returns>Returns null because I was to lazy to read into XML-Schema and implement one</returns>
		public System.Xml.Schema.XmlSchema GetSchema() {
			return null;
		}
		//==========================================================================
		#endregion
	}
}